#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#     Miguel Ángel Fernández <mafesan@bitergia.com>
#

import argparse
import logging
import sys

import redis
import rq
import rq.utils

from arthur.common import (Q_ARCHIVE_JOBS,
                           Q_CREATION_JOBS,
                           Q_RETRYING_JOBS,
                           Q_UPDATING_JOBS,
                           CH_PUBSUB)
from arthur.worker import ArthurWorker


ARTHURW_USAGE_MSG = \
    """%(prog)s [-g] [-d <database>] [--burst] [<queue1>...<queueN>] | --help """

ARTHURW_DESC_MSG = \
    """King Arthur's worker. It will run Perceval jobs on the quest
to retrieve data from software repositories.

positional arguments:
   queues               list of queues this worker will listen for
                        ('create' and 'update', by default)

optional arguments:
  -?, --help            show this help message and exit
  -g, --debug           set debug mode on
  -d, --database        URL database connection (default: 'redis://localhost/8')
  -b, --burst           Run in burst mode (quit after all work is done)
  --pubsub-channel      name of redis pubsub channel on which to publish updates
"""

# Logging formats
ARTHURW_LOG_FORMAT = "[%(asctime)s] - %(message)s"
ARTHURW_DEBUG_LOG_FORMAT = "[%(asctime)s - %(name)s - %(levelname)s] - %(message)s"


def main():
    """Run King Arthur's worker"""

    args = parse_args()

    conn = connect_to_redis(args.database)

    queues = args.queues or [Q_ARCHIVE_JOBS, Q_CREATION_JOBS, Q_RETRYING_JOBS, Q_UPDATING_JOBS]
    burst = args.burst

    configure_logging(args.debug)

    try:
        queues = [rq.Queue(queue, connection=conn) for queue in queues]

        w = ArthurWorker(queues, connection=conn)
        w.pubsub_channel = args.pubsub_channel
        w.work(burst=burst)
    except ConnectionError as e:
        print(e)
        sys.exit(1)


def parse_args():
    """Parse command line arguments"""

    parser = argparse.ArgumentParser(usage=ARTHURW_USAGE_MSG,
                                     description=ARTHURW_DESC_MSG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     add_help=False)

    parser.add_argument('-?', '--help', action='help',
                        help=argparse.SUPPRESS)
    parser.add_argument('-g', '--debug', dest='debug',
                        action='store_true',
                        help=argparse.SUPPRESS)
    parser.add_argument('-d', '--database', dest='database',
                        default='redis://localhost/8',
                        help=argparse.SUPPRESS)
    parser.add_argument('-b', '--burst', dest='burst',
                        action='store_true',
                        help=argparse.SUPPRESS)
    parser.add_argument('queues', nargs='*',
                        help=argparse.SUPPRESS)
    parser.add_argument('--pubsub-channel', dest='pubsub_channel',
                        default=CH_PUBSUB,
                        help=argparse.SUPPRESS)
    return parser.parse_args()


def configure_logging(debug=False):
    """Configure Arthur's worker logging.

    The function configures the log messages produced by Arthur
    worker and Perceval backends. By default, log messages are sent
    to stderr. Set the parameter `debug` to activate the debug
    mode.

    :param log_path: path where logs will be stored
    :param debug: set the debug mode
    """
    logger = logging.getLogger()

    if not debug:
        logger.setLevel(logging.INFO)
        formatter = logging.Formatter(fmt=ARTHURW_LOG_FORMAT)
        logging.getLogger('requests').setLevel(logging.WARNING)
        logging.getLogger('urrlib3').setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter(fmt=ARTHURW_DEBUG_LOG_FORMAT)

    handler = rq.utils.ColorizingStreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def connect_to_redis(url):
    """Create a connection with a Redis database"""

    conn = redis.StrictRedis.from_url(url)

    return conn


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        s = "\n\nReceived Ctrl-C or other break signal. Exiting.\n"
        sys.stderr.write(s)
        sys.exit(0)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#

from contextlib import contextmanager
import datetime
import json
import os
import unittest.mock

import cherrypy
import dateutil
import requests
import rq

from arthur.server import ArthurServer
from arthur.tasks import TaskStatus

from base import TestBaseRQ


@contextmanager
def run_server(conn):
    """Start CherryPy server for the tests in the background and shut it down once a test is finished"""

    server = ArthurServer(conn, None, async_mode=False, writer=None)
    cherrypy.tree.mount(server, '/')
    cherrypy.engine.start()
    cherrypy.engine.wait(cherrypy.engine.states.STARTED)
    yield server
    cherrypy.engine.exit()
    cherrypy.engine.block()


class TestArthurServer(TestBaseRQ):
    """Unit tests for ArthurServer"""

    def test_init(self):
        """Check arguments initialization"""

        with run_server(self.conn) as server:
            self.assertTrue(server.conn, self.conn)
            self.assertIsNone(server.archive_path, None)
            self.assertIsNotNone(server._tasks)
            self.assertIsNotNone(server._scheduler)

    def test_add(self):
        """Check whether tasks are added"""

        with run_server(self.conn) as server:
            self.assertEqual(len(server._tasks.tasks), 0)

            data = {
                "tasks": [
                    {
                        "task_id": "arthur.git",
                        "backend": "git",
                        "backend_args": {
                            "gitpath": os.path.join(self.dir, 'data/git_log.txt'),
                            "uri": "http://example.com/",
                        },
                        "category": "commit",
                        "archive": {},
                        "scheduler": {
                            "delay": 10
                        }
                    }
                ]
            }

            response = requests.post("http://127.0.0.1:8080/add",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)
            self.assertTrue(len(server._tasks.tasks), 1)

            t = server._tasks.tasks[0]

            self.assertEqual(t.task_id, data["tasks"][0]["task_id"])
            self.assertEqual(t.backend, data["tasks"][0]["backend"])

    def test_remove(self):
        """Check whether tasks are removed"""

        with run_server(self.conn) as server:
            data = {
                "tasks": [
                    {
                        "task_id": "arthur.git",
                        "backend": "git",
                        "backend_args": {
                            "gitpath": os.path.join(self.dir, 'data/git_log.txt'),
                            "uri": "http://example.com/",
                        },
                        "category": "commit",
                        "archive": {},
                        "scheduler": {
                            "delay": 10
                        }
                    },
                    {
                        "task_id": "bugzilla_redhat",
                        "backend": "bugzilla",
                        "backend_args": {
                            "url": "https://bugzilla.example.com/",
                            "from_date": "2016-09-19"
                        },
                        "category": "bug",
                        "archive": {
                            'archive_path': '/tmp/archive',
                            'fetch_from_archive': False
                        },
                        "scheduler": {
                            "delay": 60
                        }
                    }
                ]
            }

            response = requests.post("http://127.0.0.1:8080/add",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)
            self.assertTrue(len(server._tasks.tasks), 2)

            response = requests.post("http://127.0.0.1:8080/remove",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)
            self.assertEqual(len(server._tasks.tasks), 0)

    def test_reschedule(self):
        """Check whether tasks are rescheduled"""

        with run_server(self.conn) as server:
            backend_args = {
                "gitpath": os.path.join(self.dir, 'data/git_log.txt'),
                "uri": "http://example.com/"
            }

            task = server._tasks.add('arthur.git', 'git', 'commit',
                                     backend_args)
            server._tasks.add('arthur_extra.git', 'git', 'commit',
                              backend_args)

            # Update status to check if it is re-scheduled again
            task.status = TaskStatus.FAILED
            server._tasks.update('arthur.git', task)

            data = {
                "tasks": [
                    {
                        "task_id": "arthur.git"
                    },
                    {
                        "task_id": "arthur_extra.git",
                    }
                ]
            }

            response = requests.post("http://127.0.0.1:8080/reschedule",
                                     json=data, headers={'Content-Type': 'application/json'})
            content = json.loads(response.content.decode('utf8').replace("'", '"'))

            self.assertTrue(response.status_code, 200)
            self.assertEqual(content["tasks"]["arthur.git"], True)
            self.assertEqual(content["tasks"]["arthur_extra.git"], False)

            # Only the failed task was re-scheduled again
            task = server._tasks.get('arthur.git')
            self.assertEqual(task.status, TaskStatus.SCHEDULED)

            task = server._tasks.get('arthur_extra.git')
            self.assertEqual(task.status, TaskStatus.NEW)

    def test_tasks(self):
        """Check whether tasks are retrieved"""

        with run_server(self.conn) as server:
            data = {
                "tasks": [
                    {
                        "task_id": "arthur.git",
                        "backend": "git",
                        "backend_args": {
                            "gitpath": os.path.join(self.dir, 'data/git_log.txt'),
                            "uri": "http://example.com/",
                        },
                        "category": "commit",
                        "archive": {},
                        "scheduler": {
                            "delay": 10
                        }
                    },
                    {
                        "task_id": "bugzilla_redhat",
                        "backend": "bugzilla",
                        "backend_args": {
                            "url": "https://bugzilla.example.com/",
                            "from_date": "2016-09-19"
                        },
                        "category": "bug",
                        "archive": {
                            'archive_path': '/tmp/archive',
                            'fetch_from_archive': True,
                            'archived_after': "2010-10-10",
                        },
                        "scheduler": {
                            "delay": 60
                        }
                    }
                ]
            }

            response = requests.post("http://127.0.0.1:8080/add",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)
            self.assertTrue(len(server._tasks.tasks), 2)

            response = requests.post("http://127.0.0.1:8080/tasks",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)

            content = json.loads(response.content.decode('utf8').replace("'", '"'))

            self.assertEqual(len(content['tasks']), len(server._tasks.tasks))
            self.assertEqual(content["tasks"][0]["task_id"], data["tasks"][0]["task_id"])
            self.assertEqual(content["tasks"][1]["task_id"], data["tasks"][1]["task_id"])

    @unittest.mock.patch('arthur.tasks.datetime_utcnow')
    def test_task(self, mock_utcnow):
        """Check whether task data is retrieved"""

        mock_utcnow.return_value = datetime.datetime(2017, 1, 1,
                                                     tzinfo=dateutil.tz.tzutc())

        with run_server(self.conn) as server:
            data = {
                "tasks": [
                    {
                        "task_id": "arthur.git",
                        "backend": "git",
                        "backend_args": {
                            "gitpath": os.path.join(self.dir, 'data/git_log.txt'),
                            "uri": "http://example.com/",
                        },
                        "category": "commit",
                        "archive": {},
                        "scheduler": {
                            "delay": 10
                        }
                    }
                ]
            }
            response = requests.post("http://127.0.0.1:8080/add",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)
            self.assertTrue(len(server._tasks.tasks), 2)

            # Run the tasks
            server.start()

            response = requests.post("http://127.0.0.1:8080/task/arthur.git",
                                     json=data, headers={'Content-Type': 'application/json'})
            self.assertTrue(response.status_code, 200)

            task = json.loads(response.content.decode('utf8').replace("'", '"'))
            jobs = task.pop('jobs')

            expected_task = {
                'task_id': 'arthur.git',
                'backend': 'git',
                'status': 'ENQUEUED',
                'age': 1,
                'num_failures': 0,
                'backend_args': {
                    'gitpath': os.path.join(self.dir, 'data/git_log.txt'),
                    'uri': 'http://example.com/',
                },
                'created_on': 1483228800.0,
                'category': 'commit',
                'archiving_cfg': None,
                'scheduling_cfg': {
                    'delay': 10,
                    'max_retries': 3,
                    'max_age': None,
                    'queue': None
                }
            }

            # Check task data
            self.assertDictEqual(task, expected_task)
            self.assertEqual(len(jobs), 1)

            # Check jobs data
            job = jobs[0]
            self.assertEqual(job['task_id'], 'arthur.git')
            self.assertEqual(job['job_number'], 1)
            self.assertEqual(job['job_status'], 'finished')

            # TODO: mock job ids generator
            expected_job_result = {
                'job_number': 1,
                'task_id': 'arthur.git',
                'fetched': 9,
                'skipped': 0,
                'min_updated_on': 1344965413.0,
                'max_updated_on': 1392185439.0,
                'last_updated_on': 1344965413.0,
                'last_uuid': '1375b60d3c23ac9b81da92523e4144abc4489d4c',
                'min_offset': None,
                'max_offset': None,
                'last_offset': None,
                'extras': None
            }

            job_result = job['result']
            job_result.pop('job_id')

            self.assertEqual(job_result, expected_job_result)

    @unittest.mock.patch('arthur.tasks.datetime_utcnow')
    def test_job(self, mock_utcnow):
        """Check whether job data is retrieved"""

        mock_utcnow.return_value = datetime.datetime(2017, 1, 1,
                                                     tzinfo=dateutil.tz.tzutc())

        with run_server(self.conn) as server:
            data = {
                "tasks": [
                    {
                        "task_id": "arthur.git",
                        "backend": "git",
                        "backend_args": {
                            "gitpath": os.path.join(self.dir, 'data/git_log.txt'),
                            "uri": "http://example.com",
                        },
                        "category": "commit",
                        "archive": {},
                        "scheduler": {
                            "delay": 10
                        }
                    }
                ]
            }
            response = requests.post("http://127.0.0.1:8080/add",
                                     json=data, headers={'Content-Type': 'application/json'})

            self.assertTrue(response.status_code, 200)
            self.assertTrue(len(server._tasks.tasks), 2)

            # Run the tasks
            server.start()

            response = requests.post("http://127.0.0.1:8080/task/arthur.git",
                                     json=data, headers={'Content-Type': 'application/json'})
            self.assertTrue(response.status_code, 200)

            task = json.loads(response.content.decode('utf8').replace("'", '"'))
            jobs = task.pop('jobs')

            # Check jobs data
            job_task = jobs[0]
            self.assertEqual(job_task['task_id'], 'arthur.git')
            self.assertEqual(job_task['job_status'], 'finished')

            job_task_result = job_task['result']

            # Save log in the meta field of the job because workers don't run in sync mode/testing mode, so logs are
            # not stored into the job
            mock_logs = list()
            first_log = {
                'created': mock_utcnow.return_value.timestamp(),
                'msg': "First log",
                'module': "foo",
                'level': 0
            }
            last_log = {
                'created': mock_utcnow.return_value.timestamp(),
                'msg': "Last log",
                'module': "foo",
                'level': 0
            }
            mock_logs.append(first_log)
            mock_logs.append(last_log)

            job_rq = rq.job.Job.fetch(job_task_result['job_id'], connection=self.conn)
            job_rq.meta['log'] = mock_logs
            job_rq.save_meta()

            response = requests.post("http://127.0.0.1:8080/job/{}".format(job_task_result['job_id']),
                                     json=data, headers={'Content-Type': 'application/json'})
            self.assertTrue(response.status_code, 200)

            job = json.loads(response.content.decode('utf8'))

            self.assertEqual(job['job_id'], job_task_result['job_id'])
            self.assertEqual(job['job_number'], 1)
            self.assertEqual(job['job_status'], 'finished')
            self.assertEqual(job['timeout'], 86400)
            self.assertEqual(job['origin'], 'create')
            self.assertEqual(job['result'], job_task_result)
            self.assertEqual(job['log'], mock_logs)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import copy
import datetime
import json
import threading
import time
import unittest

from arthur.utils import RWLock, JSONEncoder


class RWLockThread(threading.Thread):

    def __init__(self, lock, rw_buffer, sleep_before, sleep_after):
        super().__init__()
        self.lock = lock
        self.rw_buffer = rw_buffer
        self.sleep_before = sleep_before
        self.sleep_after = sleep_after
        self.entry_time = None
        self.exit_time = None


class Reader(RWLockThread):

    def __init__(self, lock, rw_buffer, sleep_before, sleep_after):
        super().__init__(lock, rw_buffer, sleep_before, sleep_after)
        self.rdata = None

    def run(self):
        time.sleep(self.sleep_before)
        self.lock.reader_acquire()
        self.entry_time = time.time()
        self.rdata = copy.deepcopy(self.rw_buffer)
        time.sleep(self.sleep_after)
        self.exit_time = time.time()
        self.lock.reader_release()


class Writer(RWLockThread):

    def __init__(self, lock, rw_buffer, sleep_before, sleep_after, data):
        super().__init__(lock, rw_buffer, sleep_before, sleep_after)
        self.wdata = data

    def run(self):
        time.sleep(self.sleep_before)
        self.lock.writer_acquire()
        self.entry_time = time.time()
        self.rw_buffer[0] = self.wdata
        time.sleep(self.sleep_after)
        self.exit_time = time.time()
        self.lock.writer_release()


class TestRWLock(unittest.TestCase):
    """Unit tests for RWLock class"""

    def test_multiple_readers(self):
        """Test non-exclusive access to readers"""

        rw_lock = RWLock()
        rw_buffer = ['A']

        threads = [Reader(rw_lock, rw_buffer, 0, 0.05),
                   Reader(rw_lock, rw_buffer, 0.01, 0),
                   Writer(rw_lock, rw_buffer, 0.02, 0.02, 'Z'),
                   Reader(rw_lock, rw_buffer, 0.03, 0)]

        for th in threads:
            th.start()
        for th in threads:
            th.join()

        self.assertEqual(rw_buffer, ['Z'])
        self.assertEqual(threads[0].rdata, ['A'])
        self.assertEqual(threads[1].rdata, ['A'])
        self.assertEqual(threads[3].rdata, ['Z'])

        # Second reader started after the first one finishing
        # before it. This means there is no mutual exclusion
        # between readers.
        self.assertLess(threads[0].entry_time, threads[1].entry_time)
        self.assertGreater(threads[0].exit_time, threads[1].exit_time)

    def test_exclusion_writers(self):
        """Test if only one writer can access the resource at the same time"""

        rw_lock = RWLock()
        rw_buffer = ['A']

        threads = [Reader(rw_lock, rw_buffer, 0, 0.02),
                   Writer(rw_lock, rw_buffer, 0.02, 0.03, 'Z'),
                   Writer(rw_lock, rw_buffer, 0.03, 0, 'X'),
                   Reader(rw_lock, rw_buffer, 0.04, 0)]

        for th in threads:
            th.start()
        for th in threads:
            th.join()

        self.assertEqual(rw_buffer, ['X'])
        self.assertEqual(threads[0].rdata, ['A'])
        self.assertEqual(threads[3].rdata, ['X'])

        # First writting lasts more than the time that the second
        # writer needs to access the critical section. But, the
        # second writer will wait till the first writer ends its work.
        self.assertLess(threads[1].entry_time, threads[2].entry_time)
        self.assertLess(threads[1].exit_time, threads[2].exit_time)

        # The last thread (a reader) will wait till the last writer ends.
        self.assertLess(threads[2].entry_time, threads[3].entry_time)
        self.assertLess(threads[2].exit_time, threads[3].exit_time)


class TestJSONEncoder(unittest.TestCase):
    """Unit tests for JSONEncoder class"""

    def test_default(self):
        """Test default method"""

        encoder = JSONEncoder()

        dt = datetime.datetime(2016, 1, 1, 8, 8, 8)
        value = encoder.default(dt)
        self.assertEqual(value, "2016-01-01T08:08:08")

        # This method will raise TypeError exceptions for those
        # objects that are not instances of `datetime`
        self.assertRaises(TypeError, encoder.default, 8)
        self.assertRaises(TypeError, encoder.default, 'test')
        self.assertRaises(TypeError, encoder.default, {})
        self.assertRaises(TypeError, encoder.default, [1, 2, 3])

    def test_iterencode(self):
        """Test iterencode method"""

        encoder = JSONEncoder()

        obj = {
            'l': [None],
            'dt': datetime.datetime(2016, 1, 1, 8, 8, 8),
            's': "a string",
            'i': 8
        }

        expected = obj
        expected['dt'] = "2016-01-01T08:08:08"

        # Join the encoded JSON parts and decode the result
        # into a new object. This is needed because the joined
        # string does not follow always the same order.
        result = ''.join([chunk for chunk in encoder.iterencode(obj)])
        result = json.loads(result)
        self.assertEqual(result, obj)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#     Alvaro del Castillo San Felix <acs@bitergia.com>
#

import collections
import datetime
import enum
import logging
import re
import pickle

from redis.exceptions import RedisError

from grimoirelab_toolkit.datetime import (InvalidDateError,
                                          datetime_to_utc,
                                          datetime_utcnow,
                                          str_to_datetime)
from grimoirelab_toolkit.introspect import find_class_properties

import perceval.backend
import perceval.backends

from .common import MAX_JOB_RETRIES, WAIT_FOR_QUEUING
from .errors import (AlreadyExistsError,
                     NotFoundError,
                     TaskRegistryError)
from .utils import RWLock


logger = logging.getLogger(__name__)


TASK_PREFIX = 'arthur:task:'


@enum.unique
class TaskStatus(enum.Enum):
    """Task life cycle statuses.

    The life cycle of a task starts when is created and added
    to the system as `NEW`.

    The next step will be to program a task job. Once this is
    done the task will be `SCHEDULED` to run at a defined time.
    It will remain in this status until its job is `ENQUEUED`.

    The job will advance in the queue while other jobs are
    executed. Right after it gets to the head of the queue and a
    worker is free it will execute. The task will be `RUNNING`.

    Depending on the result executing the job, the outcomes will
    be different. If the job executed successfully, the task
    will be set to `COMPLETED`. If there was an error the status
    will be `FAILED`.

    Recurring tasks, that were successful, will be re-scheduled
    again (`SCHEDULED`), stating a new cycle.
    """
    NEW = 1
    SCHEDULED = 2
    ENQUEUED = 3
    RUNNING = 4
    COMPLETED = 5
    FAILED = 6


class Task:
    """Basic class to store information about a task.

    This class stores the basic information needed to retrieve data
    from a repository. The parameters needed to run the backend
    are given in the dictionary `backend_args`. Other parameters
    can also given to configure the archive (with `archive_args`) or to
    define how this task will be scheduled (using an instance of
    `SchedulingTaskConfig`).

    The task will be identified by the `task_id` attribute.

    :param task_id: identifier of this task
    :param backend: backend used to fetch data from the repository
    :param category: category of the items to fetch
    :param backend_args: dict of arguments required to run the backend
    :param archiving_cfg: archiving config for this task, if needed
    :param scheduling_cfg: scheduling config for this task, if needed

    :raises NotFoundError: when the given backend is not available
    """
    def __init__(self, task_id, backend, category, backend_args,
                 archiving_cfg=None, scheduling_cfg=None):
        try:
            bklass = perceval.backend.find_backends(perceval.backends)[0][backend]
        except KeyError:
            raise NotFoundError(element=backend)

        self._task_id = task_id
        self._has_resuming = bklass.has_resuming()

        self.status = TaskStatus.NEW
        self.age = 0
        self.num_failures = 0
        self.jobs = []
        self.created_on = datetime_utcnow().timestamp()
        self.backend = backend
        self.category = category
        self.backend_args = backend_args
        self.archiving_cfg = archiving_cfg if archiving_cfg else None
        self.scheduling_cfg = scheduling_cfg if scheduling_cfg else None

    @property
    def task_id(self):
        return self._task_id

    def has_resuming(self):
        """Returns if the task can be resumed when it fails"""

        return self._has_resuming

    def set_job(self, job_id, job_number):
        """Adds a JobData tuple object to the job list of a task"""

        job_tuple = JobData(job_id, job_number)
        self.jobs.append(job_tuple)

    def to_dict(self):
        return {
            'task_id': self.task_id,
            'status': self.status.name,
            'age': self.age,
            'num_failures': self.num_failures,
            'jobs': [{
                'job_id': job.id,
                'job_number': job.number
            } for job in self.jobs],
            'created_on': self.created_on,
            'backend': self.backend,
            'backend_args': self.backend_args,
            'category': self.category,
            'archiving_cfg': self.archiving_cfg.to_dict() if self.archiving_cfg else None,
            'scheduling_cfg': self.scheduling_cfg.to_dict() if self.scheduling_cfg else None
        }


JobData = collections.namedtuple('JobData', ['id', 'number'])


class TaskRegistry:
    """Structure to register tasks.

    Tasks are stored using instances of `Task` class. Each task
    is added using its tag as unique identifier. Following accesses
    to the registry (i.e, to get or remove) will require of this
    identifier.

    The registry ensures mutual exclusion among threads using a
    reading-writing lock (`RWLock` on `utils` module).

    :param conn: connection Object to Redis
    """
    def __init__(self, conn):
        self.conn = conn
        self._rwlock = RWLock()

    @staticmethod
    def _task_key(task_id):
        return '{}:{}'.format(TASK_PREFIX, task_id)

    def add(self, task_id, backend, category, backend_args,
            archiving_cfg=None, scheduling_cfg=None):
        """Add a task to the registry.

        This method adds task using `task_id` as identifier. If a task
        with the same identifier already exists on the registry, a
        `AlreadyExistsError` exception will be raised.

        :param task_id: identifier of the task to add
        :param backend: backend used to fetch data from the repository
        :param category: category of the items to fetch
        :param backend_args: dictionary of arguments required to run the backend
        :param archiving_cfg: archiving config for the task, if needed
        :param scheduling_cfg: scheduling config for the task, if needed

        :returns: the new task added to the registry

        :raises AlreadyExistsError: raised when the given task identifier
            already exists in the registry
        :raises TaskRegistryError: raised when the given task identifier
            is not added to the registry
        """
        self._rwlock.writer_acquire()
        try:
            task_key = self._task_key(task_id)
            found = self.conn.exists(task_key)

            if found:
                raise AlreadyExistsError(element=str(task_id))

            task = Task(task_id, backend, category, backend_args,
                        archiving_cfg=archiving_cfg,
                        scheduling_cfg=scheduling_cfg)
            self.conn.set(task_key, pickle.dumps(task))

            logger.debug("Task %s added to the registry", str(task_id))
            return task
        except RedisError as e:
            msg = "Task {} not added: {}".format(task_id, e)
            logger.error(msg)
            raise TaskRegistryError(cause=msg)
        finally:
            self._rwlock.writer_release()

    def remove(self, task_id):
        """Remove a task from the registry.

        To remove it, pass its identifier with `task_id` parameter.
        When the identifier is not found, a `NotFoundError` exception
        is raised.

        :param task_id: identifier of the task to remove

        :raises NotFoundError: raised when the given task identifier
            is not found in the registry
        :raises TaskRegistryError: raised when the given task identifier
            is not removed from the registry
        """
        self._rwlock.writer_acquire()
        try:
            task_key = self._task_key(task_id)
            found = self.conn.exists(task_key)

            if not found:
                raise NotFoundError(element=str(task_id))

            self.conn.delete(task_key)

            logger.debug("Task %s removed from the registry", str(task_id))
        except RedisError as e:
            msg = "Task {} not removed: {}".format(task_id, e)
            logger.error(msg)
            raise TaskRegistryError(cause=msg)
        finally:
            self._rwlock.writer_release()

    def get(self, task_id):
        """Get a task from the registry.

        Retrieve a task from the registry using its task identifier. When
        the task does not exist, a `NotFoundError` exception will be
        raised.

        :param task_id: task identifier

        :returns: a task object

        :raises NotFoundError: raised when the requested task is not
-           found in the registry
        :raises TaskRegistryError: raised when the requested task is not
            retrieved from the registry
        """
        self._rwlock.reader_acquire()
        try:
            task_key = self._task_key(task_id)
            found = self.conn.exists(task_key)

            if not found:
                raise NotFoundError(element=str(task_id))

            task_dump = self.conn.get(task_key)
            task = pickle.loads(task_dump)

            return task
        except RedisError as e:
            msg = "Task {} not retrieved: {}".format(task_id, e)
            logger.error(msg)
            raise TaskRegistryError(cause=msg)
        finally:
            self._rwlock.reader_release()

    def update(self, task_id, task):
        """Update a task in the registry.

        Update a task stored in the registry using its task identifier. When
        the task does not exist, a `NotFoundError` exception will be
        raised.

        :param task_id: task identifier
        :param task: task object

        :returns: a task object

        :raises TaskRegistryError: raised when the task is not
            updated
        """
        self._rwlock.writer_acquire()
        try:
            task_key = self._task_key(task_id)
            found = self.conn.exists(task_key)

            if not found:
                logger.warning("Task %s not found, adding it", str(task_id))

            self.conn.set(task_key, pickle.dumps(task))

            logger.debug("Task %s updated", str(task_id))
        except RedisError as e:
            msg = "Task {} not updated: {}".format(task_id, e)
            logger.error(msg)
            raise TaskRegistryError(cause=msg)
        finally:
            self._rwlock.writer_release()

    @property
    def tasks(self):
        """Get the list of tasks

        Retrieve the list of tasks stored in the registry

        :returns: a list of tasks

        :raises TaskRegistryError: raised when the tasks cannot
            be listed
        """
        self._rwlock.reader_acquire()
        try:
            tasks = []
            keys = []

            match_prefix = "{}*".format(TASK_PREFIX)
            total, found = self.conn.scan(match=match_prefix)
            keys.extend([f.decode("utf-8") for f in found])
            while total != 0:
                total, found = self.conn.scan(cursor=total, match=match_prefix)
                keys.extend([f.decode("utf-8") for f in found])

            keys.sort()
            for k in keys:
                task_dump = self.conn.get(k)
                tasks.append(pickle.loads(task_dump))

            return tasks
        except RedisError as e:
            msg = "Tasks not listed: {}".format(e)
            logger.error(msg)
            raise TaskRegistryError(cause=msg)
        finally:
            self._rwlock.reader_release()


class _TaskConfig:
    """Abstract class to store task configuration options.

    This class defines how to store specific task configuration
    arguments such as scheduling or archiving options. It is not
    meant to be instantiated on its own.

    Configuration options must be defined using `property` and `setter`
    decorators. Setters must check whether the given value is valid
    or not. When it is invalid, a `ValueError` exception should be
    raised. The rationale behind this is to use these methods as
    parsers when `from_dict` class method is called. It will create
    a new instance of the subclass passing its properties from a
    dictionary.
    """
    KW_ARGS_ERROR_REGEX = re.compile(r"^.+ got an unexpected keyword argument '(.+)'$")

    def to_dict(self):
        """Returns a dict with the representation of this task configuration object."""

        properties = find_class_properties(self.__class__)
        config = {
            name: self.__getattribute__(name) for name, _ in properties
        }
        return config

    @classmethod
    def from_dict(cls, config):
        """Create an configuration object from a dictionary.

        Key,value pairs will be used to initialize a task configuration
        object. If 'config' contains invalid configuration parameters
        a `ValueError` exception will be raised.

        :param config: dictionary used to create an instance of this object

        :returns: a task config instance

        :raises ValueError: when an invalid configuration parameter is found
        """
        try:
            obj = cls(**config)
        except TypeError as e:
            m = cls.KW_ARGS_ERROR_REGEX.match(str(e))
            if m:
                raise ValueError("unknown '%s' task config parameter" % m.group(1))
            else:
                raise e
        else:
            return obj


class ArchivingTaskConfig(_TaskConfig):
    """Manages the archiving configuration of a task.

    A limited number of archiving parameters can be configured for a
    task.

    The `archive_path` option stores the path where the archive
    for this task is or will be stored.

    The `fetch_from_archive` option sets if the task will fetch
    items from the archive.

    When `archived_after` is set, only those items archived after
    this date will be fetched.

    :param archive_path: path where the archive is or will be stored
    :param fetch_from_archive: fetch items from the archive
    :param archived_after: fetch items archived after the given date
    """
    def __init__(self, archive_path, fetch_from_archive,
                 archived_after=None):
        self.archive_path = archive_path
        self.fetch_from_archive = fetch_from_archive
        self.archived_after = archived_after

    @property
    def archive_path(self):
        """Path where the archive for this task is or will be stored."""

        return self._archive_path

    @archive_path.setter
    def archive_path(self, value):
        if not isinstance(value, str):
            raise ValueError("'archive_path' must be a str; %s given"
                             % str(type(value)))
        self._archive_path = value

    @property
    def fetch_from_archive(self):
        """Defines if the task will fetch items from the archive."""

        return self._fetch_from_archive

    @fetch_from_archive.setter
    def fetch_from_archive(self, value):
        if not isinstance(value, bool):
            raise ValueError("'fetch_from_archive' must be a bool; %s given"
                             % str(type(value)))
        self._fetch_from_archive = value

    @property
    def archived_after(self):
        """Items archived after this date will be fetched."""

        return self._archived_after

    @archived_after.setter
    def archived_after(self, value):
        if value is None:
            self._archived_after = None
        elif isinstance(value, datetime.datetime):
            self._archived_after = datetime_to_utc(value)
        elif isinstance(value, str):
            try:
                self._archived_after = str_to_datetime(value)
            except InvalidDateError as e:
                raise ValueError("'archived_after' is invalid; %s" % str(e))
        else:
            raise ValueError("'archived_after' must be either a str or a datetime; %s given"
                             % str(type(value)))


class SchedulingTaskConfig(_TaskConfig):
    """Manages the scheduling configuration of a task.

    A limited number of parameters can be configured to schedule a task.

    The `delay` option stores the number of seconds a recurring
    task will be waiting before being scheduled again.

    The `max_retries` option configures the maximum number of attempts
    a job can execute before failing.

    The `max_age` option defines the maximum number of a task can run
    in the scheduler. Set it to `None` to run the task indefinitely.

    The 'queue' option defines the name of the queue where this
    task will run. Set it to `None` to let the scheduler decide
    what the best queue is.

    :param delay: seconds of delay
    :param max_retries: maximum number of job retries before failing
    :param max_age: maximum number of times the task can run in the scheduler
    :param queue: name of the queue to run this task
    """
    def __init__(self, delay=WAIT_FOR_QUEUING, max_retries=MAX_JOB_RETRIES,
                 max_age=None, queue=None):
        self.delay = delay
        self.max_retries = max_retries
        self.max_age = max_age
        self.queue = queue

    @property
    def delay(self):
        """Number of seconds a recurring task will be waiting before being scheduled again"""

        return self._delay

    @delay.setter
    def delay(self, value):
        if not isinstance(value, int):
            raise ValueError("'delay' must be an int; %s given" % str(type(value)))
        self._delay = value

    @property
    def max_retries(self):
        """Maximum  number of attempts this job can execute before failing."""

        return self._max_retries

    @max_retries.setter
    def max_retries(self, value):
        if not isinstance(value, int):
            raise ValueError("'max_retries' must be an int; %s given" % str(type(value)))
        self._max_retries = value

    @property
    def max_age(self):
        """Maximum number of times this task can run in the scheduler.

        When it is set to `None`, the task will run indefinitely.
        """
        return self._max_age

    @max_age.setter
    def max_age(self, value):
        if value is not None:
            if not isinstance(value, int):
                raise ValueError("'max_age' must be an int; %s given" % str(type(value)))
            elif value < 1:
                raise ValueError("'max_age' must have a positive value; %s given" % str(value))
        self._max_age = value

    @property
    def queue(self):
        """Name of the queue where the task will run.

        When it is set to `None`, the scheduler will decide which queue to use.
        """
        return self._queue

    @queue.setter
    def queue(self, value):
        if value is not None:
            if not isinstance(value, str):
                raise ValueError("'queue' must be a str; %s given" % str(type(value)))
        self._queue = value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#     Alvaro del Castillo San Felix <acs@bitergia.com>
#

import copy
import logging
import os
import pickle

import rq

from .common import CH_PUBSUB, ARCHIVES_DEFAULT_PATH, Q_STORAGE_ITEMS
from .errors import AlreadyExistsError, NotFoundError, TaskRegistryError
from .scheduler import Scheduler
from .tasks import (ArchivingTaskConfig,
                    SchedulingTaskConfig,
                    TaskRegistry,
                    TaskStatus)

logger = logging.getLogger(__name__)


class Arthur:
    """Main class to retrieve data from software repositories.

    :param conn: connection Object to Redis
    :param base_archive_path: path where the archive manager is located
    :param async_mode: run in async mode (with workers); set to `False`
        for debugging purposes
    """
    def __init__(self, conn, base_archive_path=None, async_mode=True, pubsub_channel=CH_PUBSUB):
        self.conn = conn
        self.conn.flushdb()
        rq.push_connection(self.conn)

        self.archive_path = base_archive_path
        self._tasks = TaskRegistry(self.conn)
        self._scheduler = Scheduler(self.conn, self._tasks,
                                    pubsub_channel=pubsub_channel,
                                    async_mode=async_mode)

    def start(self):
        self._scheduler.schedule()

    def add_task(self, task_id, backend, category, backend_args,
                 archive_args=None, sched_args=None):
        """Add and schedule a task.

        :param task_id: id of the task
        :param backend: name of the backend
        :param category: category of the items to fetch
        :param backend_args: args needed to initialize the backend
        :param archive_args: args needed to initialize the archive
        :param sched_args: scheduling args for this task

        :returns: the task created
        """
        try:
            archiving_cfg = self.__parse_archive_args(archive_args)
            scheduling_cfg = self.__parse_schedule_args(sched_args)
            self.__validate_args(task_id, backend, category, backend_args)
        except ValueError as e:
            raise e

        try:
            task = self._tasks.add(task_id, backend, category, backend_args,
                                   archiving_cfg=archiving_cfg,
                                   scheduling_cfg=scheduling_cfg)
        except AlreadyExistsError as e:
            raise e
        except TaskRegistryError as e:
            raise e

        self._scheduler.schedule_task(task.task_id)

        return task

    def remove_task(self, task_id):
        """Remove and cancel a task.

        :param task_id: id of the task to be removed
        """
        try:
            self._scheduler.cancel_task(task_id)
        except NotFoundError:
            logger.info("Cannot cancel %s task because it does not exist.",
                        task_id)
            return False
        else:
            return True

    def reschedule_task(self, task_id):
        """Re-schedule a failed task.

        :param task_id: id of the task
        """
        try:
            task = self._tasks.get(task_id)
        except NotFoundError as e:
            logger.info("Cannot re-schedule %s task because it does not exist.",
                        task_id)
            return False

        if task.status == TaskStatus.FAILED:
            self._scheduler.schedule_task(task_id, reset=True)
            return True
        else:
            logger.info("Cannot re-schedule task %s; only FAILED tasks can be rescheduled.",
                        task_id)
            return False

    def items(self):
        """Get the items fetched by the jobs."""

        # Get and remove queued items in an atomic transaction
        pipe = self.conn.pipeline()
        pipe.lrange(Q_STORAGE_ITEMS, 0, -1)
        pipe.ltrim(Q_STORAGE_ITEMS, 1, 0)
        items = pipe.execute()[0]

        for item in items:
            item = pickle.loads(item)
            yield item

    @staticmethod
    def __validate_args(task_id, backend, category, backend_args):
        """Check that the task arguments received are valid"""

        if not task_id or task_id.strip() == "":
            msg = "Missing task_id for task"
            raise ValueError(msg)

        if not backend or backend.strip() == "":
            msg = "Missing backend for task '%s'" % task_id
            raise ValueError(msg)

        if backend_args and not isinstance(backend_args, dict):
            msg = "Backend_args is not a dict, task '%s'" % task_id
            raise ValueError(msg)

        if not category or category.strip() == "":
            msg = "Missing category for task '%s'" % task_id
            raise ValueError(msg)

    def __parse_archive_args(self, archive_args):
        """Parse the archive arguments of a task"""

        if not archive_args:
            return None

        archiving_args = copy.deepcopy(archive_args)

        if self.archive_path:
            archiving_args['archive_path'] = self.archive_path
        else:
            archiving_args['archive_path'] = os.path.expanduser(ARCHIVES_DEFAULT_PATH)

        return ArchivingTaskConfig.from_dict(archiving_args)

    def __parse_schedule_args(self, sched_args):
        """Parse the schedule arguments of a task"""

        if not sched_args:
            return None

        return SchedulingTaskConfig.from_dict(sched_args)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import os.path
import unittest

from fakeredis import FakeStrictRedis
from rq import pop_connection, push_connection


# The next code is based on  'rq/tests/__init__.py' file of
# rq project. This code was licensed as BSD-like. All credits
# to their authors.

def find_empty_redis_database():
    """Connect to a fake Redis database"""
    conn = FakeStrictRedis()
    return conn


class TestBaseRQ(unittest.TestCase):
    """Base class to inherit test cases from for RQ"""

    @classmethod
    def setUpClass(cls):
        cls.conn = find_empty_redis_database()
        push_connection(cls.conn)
        cls.dir = os.path.dirname(os.path.realpath(__file__))

    @classmethod
    def tearDownClass(cls):
        conn = pop_connection()
        assert conn == cls.conn, \
            "Wow, something really nasty happened to the FakeRedis connection stack. Check your setup."

    def setUp(self):
        self.conn.flushdb()

    def tearDown(self):
        self.conn.flushdb()


def mock_sum(task_id, a, b):
    return a + b


def mock_failure(task_id):
    raise Exception



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#     Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

import rq
import rq.job

from .common import CH_PUBSUB
from .events import JobEventType, JobEvent

JOB_LOGGERS = ['arthur', 'perceval', 'rq']
logger = logging.getLogger(__name__)


class JobLogHandler(logging.StreamHandler):
    """Handler class for the job logs"""

    def __init__(self, job):
        logging.StreamHandler.__init__(self)
        self.job = job
        self.job.meta['log'] = []
        self.job.save_meta()

    def emit(self, record):
        log = {
            'created': record.created,
            'msg': self.format(record),
            'module': record.module,
            'level': self.level
        }
        self.job.meta['log'].append(log)
        self.job.save_meta()


class ArthurWorker(rq.Worker):
    """Worker class for Arthur"""

    def __init__(self, queues, **kwargs):
        super().__init__(queues, **kwargs)
        self.__pubsub_channel = CH_PUBSUB

    @property
    def pubsub_channel(self):
        return self.__pubsub_channel

    @pubsub_channel.setter
    def pubsub_channel(self, value):
        self.__pubsub_channel = value

    def perform_job(self, job, queue, heartbeat_ttl=None):
        """Custom method to execute a job and notify of its result

        :param job: Job object
        :param queue: the queue containing the object
        :param heartbeat_ttl: time to live heartbeat
        """
        self.setup_job_loghandlers(job)
        self._publish_job_event_when_started(job)
        result = super().perform_job(job, queue, heartbeat_ttl=heartbeat_ttl)
        self._publish_job_event_when_finished(job)

        return result

    def _publish_job_event_when_started(self, job):
        """Send event notifying the job started"""

        task_id = job.kwargs['task_id']

        event = JobEvent(JobEventType.STARTED, job.id, task_id,
                         None)

        msg = event.serialize()
        self.connection.publish(self.pubsub_channel, msg)

    def _publish_job_event_when_finished(self, job):
        """Send event notifying the result of a finished job"""

        job_status = job.get_status()

        if job_status == rq.job.JobStatus.FINISHED:
            event_type = JobEventType.COMPLETED
            payload = job.return_value
        elif job_status == rq.job.JobStatus.FAILED:
            event_type = JobEventType.FAILURE
            payload = {
                'error': job.exc_info,
                'result': job.meta.get('result', None)
            }
        else:
            logger.warning("Unexpected job status %s for finished job %s",
                           job_status, job.id)
            event_type = JobEventType.UNDEFINED
            payload = job_status

        task_id = job.kwargs['task_id']

        event = JobEvent(event_type, job.id, task_id, payload)

        msg = event.serialize()
        self.connection.publish(self.pubsub_channel, msg)

    def setup_job_loghandlers(self, job):
        meta_handler = JobLogHandler(job)
        for logger_name in JOB_LOGGERS:
            logger_job = logging.getLogger(logger_name)
            logger_job.addHandler(meta_handler)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import enum
import logging
import pickle
import traceback
import threading
import uuid

from grimoirelab_toolkit.datetime import datetime_utcnow

from .common import CH_PUBSUB


logger = logging.getLogger(__name__)


@enum.unique
class JobEventType(enum.Enum):
    STARTED = 1
    COMPLETED = 2
    FAILURE = 3
    UNDEFINED = 999


class JobEvent:
    """Job activity notification.

    Job events can be used to notify the activity of a job. Usually,
    they will inform whether the job was completed or finished with
    a failure.

    Each event has a type, a unique identifier and the time when it
    was generated (in UTC), the identifier of the job and task that
    produced it and a payload. Depending on the type of event, the
    payload might contain different data.

    :param type: event type
    :param job_id: identifier of the job
    :param task_id: identifier of the task
    :param payload: data of the event
    """
    def __init__(self, type, job_id, task_id, payload):
        self.uuid = str(uuid.uuid4())
        self.timestamp = datetime_utcnow()
        self.type = type
        self.job_id = job_id
        self.task_id = task_id
        self.payload = payload

    def serialize(self):
        return pickle.dumps(self)

    @classmethod
    def deserialize(cls, data):
        return pickle.loads(data)


class JobEventsListener(threading.Thread):
    """Listen jobs events.

    This class listens in a separate thread for those events produced
    by the running jobs. Events can be processed using a set of
    handlers.

    Handlers can be registered for a certain type of events using the
    method `subscribe`. To stop handling them call the method
    `unsubscribe`.

    :param conn: connection to the Redis database
    :param events_channel: name of the events channel where jobs write events
    """
    def __init__(self, conn, events_channel=CH_PUBSUB):
        super().__init__()
        self.conn = conn
        self.events_channel = events_channel
        self.handlers = {}

    def run(self):
        """Start a thread to listen for jobs events."""

        try:
            self.listen()
        except Exception as e:
            logger.critical("JobEventsListener instance crashed. Error: %s", str(e))
            logger.critical(traceback.format_exc())

    def subscribe(self, event_type, handler):
        """Activate the listener to handle events of a type.

        The listener will start processing events of the given type.
        When an event of that type is caught, the listener will call
        to `handler` passing it that event.

        Next calls to this method will replace the handler that was
        processing those events.

        :param event_type: type of event to subscribe
        :param handler: callable object to process the events

        :raises TypeError: when `even_type` is not a `JobEventType` object.
        """
        if not isinstance(event_type, JobEventType):
            msg = "'%s' object is not a JobEventType" % event_type.__class__.__name__
            raise TypeError(msg)

        self.handlers[event_type] = handler

    def unsubscribe(self, event_type):
        """Deactivate the listener to handle events of a type.

        When this method is called, the listener will stop processing
        events of the given type.

        :param event_type: type of the event to unsubscribe

        :raises TypeError: when `even_type` is not a `JobEventType` object.
        """
        if not isinstance(event_type, JobEventType):
            msg = "'%s' object is not a JobEventType" % event_type.__class__.__name__
            raise TypeError(msg)

        self.handlers[event_type] = None

    def listen(self):
        """Listen for events.

        The object will start listening for those events produced in
        the system. When an event arrives, the event will be handled
        if the listener was previously subscribed to it.

        Take into account this is a blocking call. The listener will
        be waiting for new events and it will not return the control
        to the caller.
        """
        pubsub = self.conn.pubsub()
        pubsub.subscribe(self.events_channel)

        logger.debug("Listening on channel %s", self.events_channel)

        # Redis 'listen' method is a blocking call
        for msg in pubsub.listen():
            logger.debug("New message received of type %s", str(msg['type']))

            if msg['type'] != 'message':
                logger.debug("Ignoring job message")
                continue

            event = JobEvent.deserialize(msg['data'])

            self._dispatch_event(event)

        logger.debug("End listening on channel %s", self.events_channel)

    def _dispatch_event(self, event):
        """Dispatches the job event to the right handler."""

        handler = self.handlers.get(event.type, None)

        if handler:
            logger.debug("Calling handler for job #%s", event.job_id)
            handler(event)
        else:
            logger.debug("No handler defined for %s", event.type.name)
